@@ -32,6 +32,16 @@ from nums.core.systems.systems import SerialSystem, RaySystem
 
 
 _instance: ArrayApplication = None
+_call_on_create: list = []
+
+
+def call_on_create(func):
+    global _call_on_create
+    # Always include funcs in _call_on_create.
+    # If the app is destroyed, the hooks need to be invoked again on creation.
+    _call_on_create.append(func)
+    if is_initialized():
+        func(_instance)
 
 
 def is_initialized():
@@ -43,6 +53,8 @@ def instance():
     global _instance
     if _instance is None:
         _instance = create()
+        for func in _call_on_create:
+            func(_instance)
     return _instance
 
 
@@ -57,24 +69,27 @@ def create():
     # Initialize compute interface and system.
     system_name = settings.system_name
     if system_name == "serial":
-        system: SystemInterface = SerialSystem()
+        system: SystemInterface = SerialSystem(settings.num_cpus)
     elif system_name == "ray":
         use_head = settings.use_head
         num_nodes = int(np.product(settings.cluster_shape))
-        system: SystemInterface = RaySystem(use_head=use_head,
-                                            num_nodes=num_nodes)
+        system: SystemInterface = RaySystem(
+            use_head=use_head, num_nodes=num_nodes, num_cpus=settings.num_cpus
+        )
     else:
         raise Exception("Unexpected system name %s" % settings.system_name)
     system.init()
 
-    compute_module = {
-        "numpy": numpy_compute
-    }[settings.compute_name]
+    compute_module = {"numpy": numpy_compute}[settings.compute_name]
 
     if settings.device_grid_name == "cyclic":
-        device_grid: DeviceGrid = CyclicDeviceGrid(settings.cluster_shape, "cpu", system.devices())
+        device_grid: DeviceGrid = CyclicDeviceGrid(
+            settings.cluster_shape, "cpu", system.devices()
+        )
     elif settings.device_grid_name == "packed":
-        device_grid: DeviceGrid = PackedDeviceGrid(settings.cluster_shape, "cpu", system.devices())
+        device_grid: DeviceGrid = PackedDeviceGrid(
+            settings.cluster_shape, "cpu", system.devices()
+        )
     else:
         raise Exception("Unexpected device grid name %s" % settings.device_grid_name)
 
