@@ -14,9 +14,10 @@
 # limitations under the License.
 
 
+import warnings
 import logging
 from types import FunctionType
-from typing import Any, Union, List, Dict
+from typing import Any, Union, List, Dict, Optional
 
 import ray
 
@@ -26,10 +27,14 @@ from nums.core.systems.utils import get_private_ip, get_num_cores
 from nums.core import settings
 
 
-class SerialSystem(SystemInterface):
+# pylint: disable = unused-argument
 
-    def __init__(self):
+
+class SerialSystem(SystemInterface):
+    def __init__(self, num_cpus: Optional[int] = None):
+        self.num_cpus = int(get_num_cores()) if num_cpus is None else num_cpus
         self._remote_functions: dict = {}
+        self._actors: dict = {}
 
     def init(self):
         pass
@@ -59,8 +64,18 @@ class SerialSystem(SystemInterface):
     def call(self, name: str, args, kwargs, device_id: DeviceID, options: Dict):
         return self._remote_functions[name](*args, **kwargs)
 
-    def num_cores_total(self):
-        return int(get_num_cores())
+    def register_actor(self, name: str, cls: type):
+        assert name not in self._actors
+        self._actors[name] = cls
+
+    def make_actor(self, name: str, *args, device_id: DeviceID = None, **kwargs):
+        return self._actors[name](*args, **kwargs)
+
+    def call_actor_method(self, actor, method: str, *args, **kwargs):
+        return getattr(actor, method)(*args, **kwargs)
+
+    def num_cores_total(self) -> int:
+        return self.num_cpus
 
 
 class RaySystem(SystemInterface):
@@ -69,11 +84,19 @@ class RaySystem(SystemInterface):
     Implements SystemInterface for Ray.
     """
 
-    def __init__(self, use_head=False, num_nodes=None):
+    def __init__(
+        self,
+        use_head: bool = False,
+        num_nodes: Optional[int] = None,
+        num_cpus: Optional[int] = None,
+    ):
         self._use_head = use_head
         self._num_nodes = num_nodes
+        self.num_cpus = int(get_num_cores()) if num_cpus is None else num_cpus
         self._manage_ray = True
         self._remote_functions = {}
+        self._actors: dict = {}
+        self._actor_node_index = 0
         self._available_nodes = []
         self._head_node = None
         self._worker_nodes = []
@@ -84,7 +107,7 @@ class RaySystem(SystemInterface):
         if ray.is_initialized():
             self._manage_ray = False
         if self._manage_ray:
-            ray.init()
+            ray.init(num_cpus=self.num_cpus)
         # Compute available nodes, based on CPU resource.
         if settings.head_ip is None:
             # TODO (hme): Have this be a class argument vs. using what's set in settings directly.
@@ -106,9 +129,11 @@ class RaySystem(SystemInterface):
                 self._available_nodes.append(node)
         if self._head_node is None:
             if self._use_head:
-                logging.getLogger(__name__).warning("Failed to determine which node is the head."
-                                                    " The head node will be used even though"
-                                                    " nums.core.settings.use_head = False.")
+                logging.getLogger(__name__).warning(
+                    "Failed to determine which node is the head."
+                    " The head node will be used even though"
+                    " nums.core.settings.use_head = False."
+                )
         elif self._use_head and self._has_cpu_resources(self._head_node):
             total_cpus += self._head_node["Resources"]["CPU"]
             self._available_nodes.append(self._head_node)
@@ -128,18 +153,18 @@ class RaySystem(SystemInterface):
             self._devices.append(did)
             self._device_to_node[did] = node
 
-    def _has_cpu_resources(self, node):
+    def _has_cpu_resources(self, node: dict) -> bool:
         return self._node_cpu_resources(node) > 0.0
 
-    def _node_cpu_resources(self, node):
+    def _node_cpu_resources(self, node: dict) -> float:
         return node["Resources"]["CPU"] if "CPU" in node["Resources"] else 0.0
 
-    def _node_key(self, node):
+    def _node_key(self, node: dict) -> str:
         node_key = list(filter(lambda key: "node" in key, node["Resources"].keys()))
         assert len(node_key) == 1
         return node_key[0]
 
-    def _node_ip(self, node):
+    def _node_ip(self, node: dict) -> str:
         return self._node_key(node).split(":")[1]
 
     def shutdown(self):
@@ -154,6 +179,7 @@ class RaySystem(SystemInterface):
             def warmup_func(n):
                 # pylint: disable=import-outside-toplevel
                 import random
+
                 r = ray.remote(num_cpus=1)(lambda x, y: x + y).remote
                 for _ in range(n):
                     _a = random.randint(0, 1000)
@@ -162,7 +188,7 @@ class RaySystem(SystemInterface):
 
             warmup_func(n)
 
-    def put(self, value):
+    def put(self, value: Any):
         return ray.put(value)
 
     def get(self, object_ids):
@@ -183,22 +209,48 @@ class RaySystem(SystemInterface):
             node_key = self._node_key(node)
             if "resources" in options:
                 assert node_key not in options
-            options["resources"] = {node_key: 1.0/10**4}
+            options["resources"] = {node_key: 1.0 / 10 ** 4}
         return self._remote_functions[name].options(**options).remote(*args, **kwargs)
 
-    def devices(self):
+    def devices(self) -> List[DeviceID]:
         return self._devices
 
-    def num_cores_total(self):
-        num_cores = sum(map(lambda n: n["Resources"]["CPU"], self._device_to_node.values()))
+    def num_cores_total(self) -> int:
+        num_cores = sum(
+            map(lambda n: n["Resources"]["CPU"], self._device_to_node.values())
+        )
         return int(num_cores)
 
+    def register_actor(self, name: str, cls: type):
+        if name in self._actors:
+            warnings.warn(
+                "Actor %s has already been registered. "
+                "Overwriting with %s." % (name, cls.__name__)
+            )
+            return
+        self._actors[name] = ray.remote(cls)
+
+    def make_actor(self, name: str, *args, device_id: DeviceID = None, **kwargs):
+        # Distribute actors round-robin over devices.
+        if device_id is None:
+            device_id = self._devices[self._actor_node_index]
+            self._actor_node_index = (self._actor_node_index + 1) % len(self._devices)
+        actor = self._actors[name]
+        node = self._device_to_node[device_id]
+        node_key = self._node_key(node)
+        options = {"resources": {node_key: 1.0 / 10 ** 4}}
+        return actor.options(**options).remote(*args, **kwargs)
+
+    def call_actor_method(self, actor, method: str, *args, **kwargs):
+        return getattr(actor, method).remote(*args, **kwargs)
+
 
 class RaySystemStockScheduler(RaySystem):
     """
     An implementation of the Ray system which ignores scheduling commands given
     by the caller. For testing only.
     """
+
     def call(self, name: str, args, kwargs, device_id: DeviceID, options: Dict):
         if device_id is not None:
             node = self._device_to_node[device_id]
